@@ -32,6 +32,30 @@ extern "C" {
 #include <groonga.h>
 #include "mrn_sys.h"
 
+#if (MYSQL_VERSION_ID >= 50603) || \
+    (MYSQL_VERSION_ID >= 50513 && MYSQL_VERSION_ID < 50600) || \
+    (MYSQL_VERSION_ID >= 50158 && MYSQL_VERSION_ID < 50500)
+#  define MRN_HANDLER_CLONE_NEED_NAME 1
+#endif
+
+#if MYSQL_VERSION_ID >= 50600
+#  define MRN_HANDLER_HAVE_HA_CLOSE 1
+#  define MRN_HANDLER_HAVE_HA_RND_NEXT 1
+#  define MRN_HANDLER_HAVE_HA_RND_POS 1
+#  define MRN_HANDLER_HAVE_HA_INDEX_READ_IDX_MAP 1
+#  define MRN_HANDLER_HAVE_HA_INDEX_NEXT 1
+#  define MRN_HANDLER_HAVE_HA_INDEX_PREV 1
+#  define MRN_HANDLER_HAVE_HA_INDEX_FIRST 1
+#  define MRN_HANDLER_HAVE_HA_INDEX_LAST 1
+#  define MRN_HANDLER_HAVE_HA_INDEX_NEXT_SAME 1
+
+#  define MRN_HANDLER_HAVE_MULTI_RANGE_READ 1
+#endif
+
+#if MYSQL_VERSION_ID < 50600
+  typedef Item COND;
+#endif
+
 class ha_mroonga;
 
 /* structs */
@@ -127,7 +151,9 @@ public:
 
   int create(const char *name, TABLE *form, HA_CREATE_INFO *info); // required
   int open(const char *name, int mode, uint test_if_locked);       // required
+#ifndef MRN_HANDLER_HAVE_HA_CLOSE
   int close();                                                     // required
+#endif
   int info(uint flag);                                             // required
 
   uint lock_count() const;
@@ -138,8 +164,12 @@ public:
 
   int rnd_init(bool scan);                                         // required
   int rnd_end();
+#ifndef MRN_HANDLER_HAVE_HA_RND_NEXT
   int rnd_next(uchar *buf);                                        // required
+#endif
+#ifndef MRN_HANDLER_HAVE_HA_RND_POS
   int rnd_pos(uchar *buf, uchar *pos);                             // required
+#endif
   void position(const uchar *record);                              // required
   int extra(enum ha_extra_function operation);
   int extra_opt(enum ha_extra_function operation, ulong cache_size);
@@ -162,10 +192,18 @@ public:
                      enum ha_rkey_function find_flag);
   int index_read_last_map(uchar *buf, const uchar *key,
                           key_part_map keypart_map);
-  int index_next(uchar * buf);
-  int index_prev(uchar * buf);
-  int index_first(uchar * buf);
-  int index_last(uchar * buf);
+#ifndef MRN_HANDLER_HAVE_HA_INDEX_NEXT
+  int index_next(uchar *buf);
+#endif
+#ifndef MRN_HANDLER_HAVE_HA_INDEX_PREV
+  int index_prev(uchar *buf);
+#endif
+#ifndef MRN_HANDLER_HAVE_HA_INDEX_FIRST
+  int index_first(uchar *buf);
+#endif
+#ifndef MRN_HANDLER_HAVE_HA_INDEX_LAST
+  int index_last(uchar *buf);
+#endif
   int index_next_same(uchar *buf, const uchar *key, uint keylen);
 
   int read_range_first(const key_range *start_key,
@@ -178,25 +216,38 @@ public:
   FT_INFO *ft_init_ext(uint flags, uint inx, String *key);
   int ft_read(uchar *buf);
 
-  const COND *cond_push(const COND *cond);
+  const Item *cond_push(const Item *cond);
   void cond_pop();
 
   bool get_error_message(int error, String *buf);
 
   int reset();
 
-#if MYSQL_VERSION_ID >= 50513
+#ifdef MRN_HANDLER_CLONE_NEED_NAME
   handler *clone(const char *name, MEM_ROOT *mem_root);
 #else
   handler *clone(MEM_ROOT *mem_root);
 #endif
   uint8 table_cache_type();
+#ifdef MRN_HANDLER_HAVE_MULTI_RANGE_READ
+  ha_rows multi_range_read_info_const(uint keyno, RANGE_SEQ_IF *seq,
+                                      void *seq_init_param,
+                                      uint n_ranges, uint *bufsz,
+                                      uint *flags, COST_VECT *cost);
+  ha_rows multi_range_read_info(uint keyno, uint n_ranges, uint keys,
+                                uint *bufsz, uint *flags, COST_VECT *cost);
+  int multi_range_read_init(RANGE_SEQ_IF *seq, void *seq_init_param,
+                            uint n_ranges, uint mode,
+                            HANDLER_BUFFER *buf);
+  int multi_range_read_next(char **range_info);
+#else // MRN_HANDLER_HAVE_MULTI_RANGE_READ
   int read_multi_range_first(KEY_MULTI_RANGE **found_range_p,
                              KEY_MULTI_RANGE *ranges,
                              uint range_count,
                              bool sorted,
                              HANDLER_BUFFER *buffer);
   int read_multi_range_next(KEY_MULTI_RANGE **found_range_p);
+#endif // MRN_HANDLER_HAVE_MULTI_RANGE_READ
   void start_bulk_insert(ha_rows rows);
   int end_bulk_insert();
   int delete_all_rows();
@@ -220,7 +271,30 @@ public:
   int optimize(THD* thd, HA_CHECK_OPT* check_opt);
   bool is_fatal_error(int error_num, uint flags);
 
+protected:
+#ifdef MRN_HANDLER_HAVE_HA_RND_NEXT
+  int rnd_next(uchar *buf);
+#endif
+#ifdef MRN_HANDLER_HAVE_HA_RND_POS
+  int rnd_pos(uchar *buf, uchar *pos);
+#endif
+#ifdef MRN_HANDLER_HAVE_HA_INDEX_NEXT
+  int index_next(uchar *buf);
+#endif
+#ifdef MRN_HANDLER_HAVE_HA_INDEX_PREV
+  int index_prev(uchar *buf);
+#endif
+#ifdef MRN_HANDLER_HAVE_HA_INDEX_FIRST
+  int index_first(uchar *buf);
+#endif
+#ifdef MRN_HANDLER_HAVE_HA_INDEX_LAST
+  int index_last(uchar *buf);
+#endif
+
 private:
+#ifdef MRN_HANDLER_HAVE_HA_CLOSE
+  int close();
+#endif
   void check_count_skip(key_part_map start_key_part_map,
                         key_part_map end_key_part_map, bool fulltext);
   void check_fast_order_limit();
@@ -349,13 +423,13 @@ private:
   FT_INFO *storage_ft_init_ext(uint flags, uint key_nr, String *key);
   int wrapper_ft_read(uchar *buf);
   int storage_ft_read(uchar *buf);
-  const COND *wrapper_cond_push(const COND *cond);
-  const COND *storage_cond_push(const COND *cond);
+  const Item *wrapper_cond_push(const Item *cond);
+  const Item *storage_cond_push(const Item *cond);
   void wrapper_cond_pop();
   void storage_cond_pop();
   bool wrapper_get_error_message(int error, String *buf);
   bool storage_get_error_message(int error, String *buf);
-#if MYSQL_VERSION_ID >= 50513
+#ifdef MRN_HANDLER_CLONE_NEED_NAME
   handler *wrapper_clone(const char *name, MEM_ROOT *mem_root);
   handler *storage_clone(const char *name, MEM_ROOT *mem_root);
 #else
@@ -364,6 +438,36 @@ private:
 #endif
   uint8 wrapper_table_cache_type();
   uint8 storage_table_cache_type();
+#ifdef MRN_HANDLER_HAVE_MULTI_RANGE_READ
+  ha_rows wrapper_multi_range_read_info_const(uint keyno,
+                                              RANGE_SEQ_IF *seq,
+                                              void *seq_init_param,
+                                              uint n_ranges,
+                                              uint *bufsz,
+                                              uint *flags,
+                                              COST_VECT *cost);
+  ha_rows storage_multi_range_read_info_const(uint keyno,
+                                              RANGE_SEQ_IF *seq,
+                                              void *seq_init_param,
+                                              uint n_ranges,
+                                              uint *bufsz,
+                                              uint *flags,
+                                              COST_VECT *cost);
+  ha_rows wrapper_multi_range_read_info(uint keyno, uint n_ranges, uint keys,
+                                        uint *bufsz, uint *flags,
+                                        COST_VECT *cost);
+  ha_rows storage_multi_range_read_info(uint keyno, uint n_ranges, uint keys,
+                                        uint *bufsz, uint *flags,
+                                        COST_VECT *cost);
+  int wrapper_multi_range_read_init(RANGE_SEQ_IF *seq, void *seq_init_param,
+                                    uint n_ranges, uint mode,
+                                    HANDLER_BUFFER *buf);
+  int storage_multi_range_read_init(RANGE_SEQ_IF *seq, void *seq_init_param,
+                                    uint n_ranges, uint mode,
+                                    HANDLER_BUFFER *buf);
+  int wrapper_multi_range_read_next(char **range_info);
+  int storage_multi_range_read_next(char **range_info);
+#else // MRN_HANDLER_HAVE_MULTI_RANGE_READ
   int wrapper_read_multi_range_first(KEY_MULTI_RANGE **found_range_p,
                                      KEY_MULTI_RANGE *ranges,
                                      uint range_count,
@@ -376,6 +480,7 @@ private:
                                      HANDLER_BUFFER *buffer);
   int wrapper_read_multi_range_next(KEY_MULTI_RANGE **found_range_p);
   int storage_read_multi_range_next(KEY_MULTI_RANGE **found_range_p);
+#endif // MRN_HANDLER_HAVE_MULTI_RANGE_READ
   int wrapper_delete_all_rows();
   int storage_delete_all_rows();
 #if MYSQL_VERSION_ID >= 50500
