@@ -32,10 +32,10 @@
 #define MYSQL_SERVER 1
 #include "mysql_version.h"
 
-#ifdef MYSQL51
+#if MYSQL_VERSION_ID < 50500
 #  include <mysql_priv.h>
 #  include <mysql/plugin.h>
-#else /* MYSQL51 */
+#else
 #  include <sql_priv.h>
 #  include <sql_class.h>
 #  include <probes_mysql.h>
@@ -47,6 +47,12 @@
 #include "mrn_sys.h"
 #include "mrn_table.h"
 
+#if MYSQL_VERSION_ID >= 50603
+#  define MRN_HA_RESOLVE_BY_NAME(name) ha_resolve_by_name(NULL, (name), TRUE)
+#else
+#  define MRN_HA_RESOLVE_BY_NAME(name) ha_resolve_by_name(NULL, (name))
+#endif
+
 #define MRN_DEFAULT_STR "DEFAULT"
 #define MRN_DEFAULT_LEN (sizeof(MRN_DEFAULT_STR) - 1)
 #define MRN_GROONGA_STR "GROONGA"
@@ -420,7 +426,7 @@ int mrn_parse_table_param(MRN_SHARE *share, TABLE *table)
     } else {
       engine_name.str = share->engine;
       engine_name.length = share->engine_length;
-      if (!(share->plugin = ha_resolve_by_name(NULL, &engine_name)))
+      if (!(share->plugin = MRN_HA_RESOLVE_BY_NAME(&engine_name)))
       {
         my_error(ER_UNKNOWN_STORAGE_ENGINE, MYF(0), share->engine);
         error = ER_UNKNOWN_STORAGE_ENGINE;
@@ -571,7 +577,6 @@ int mrn_free_share(MRN_SHARE *share)
 
 TABLE_SHARE *mrn_get_table_share(TABLE_LIST *table_list, int *error)
 {
-  char key[MAX_DBKEY_LENGTH];
   uint key_length;
   TABLE_SHARE *share;
 #if MYSQL_VERSION_ID >= 50500
@@ -579,7 +584,13 @@ TABLE_SHARE *mrn_get_table_share(TABLE_LIST *table_list, int *error)
 #endif
   THD *thd = current_thd;
   DBUG_ENTER("mrn_get_table_share");
+#if MYSQL_VERSION_ID >= 50603
+  const char *key;
+  key_length = get_table_def_key(table_list, &key);
+#else
+  char key[MAX_DBKEY_LENGTH];
   key_length = create_table_def_key(thd, key, table_list, FALSE);
+#endif
 #if MYSQL_VERSION_ID >= 50500
   hash_value = my_calc_hash(&table_def_cache, (uchar*) key, key_length);
   share = get_table_share(thd, table_list, key, key_length, 0, error,
